(**
 dns.ml --- host command in Ocaml
                                                                         
 Copyright (C) 2002  Guillaume Valadon <guillaume@valadon.net>
                     Manu Letellier <manu.letellier@gmail.com>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

*)

(** Easily construct dns packet. *)

(** Will be raise if we find an unknown rrtype in a packet. *)
exception Unknow_type of string;;
(** Will be raise if a timeout occurs. *)
exception TimedOut;;
(** Will be raise if a packet is malformed (for example, too small). *)
exception Malformed of string;;

(** Type that represents the data type, only internet data (IN) is 
    really usefull. *)
type classtype = 
   | IN (* 1 *)
   | CS (* 2 *)
   | CH (* 3 *)
   | HS (* 4 *)
;;

(** Type that represents ressource records. *)
type rrtype = 
   | A of string
   | AAAA of string
   | AXFR
   | CNAME of string
   | HINFO of string * string
   | MX of int * string
   | NS of string
   | NULL of string (* Anything ... *)
   | PTR of string
   | SOA of string * string * int * int * int * int * int
   | TXT of string
   | Error
   | NotFound
;;

(** Node of a N-tree, that represent the dns name space. *)
type dnstree = DomainName of string * dnstree list * rrtype list;;

(** Construct a 'dns tree' easily. *)
class dns_auth =
object (self)

   val mutable root = []
  
   (** Set the new root of the dns tree. *)
   method set aroot = root <- aroot

   (** Clear the dns tree. *)
   method clear = root <- []
   
   (** Get {i str} ressource record. *)
   method get (str:string) =
      let rec temp root strlst =
      match root with
      | DomainName (str, domainlst, rrlst)::l
      when [str] = strlst ->
	 rrlst
      | DomainName (str, domainlst, rrlst)::l
      when str = (List.hd strlst) ->
	 temp domainlst (List.tl strlst)
      | DomainName (_,_,_)::l ->
	 temp l strlst
      | [] -> 
	 [NotFound]
      in
      temp root (self#str2list str)

   (** Get {i str} subtree (very usefull for zone transfer). *)
   method get_subtree (str:string) =
      let rec temp root strlst =
      match root with
      | DomainName (str, domainlst, rrlst)::l
      when [str] = strlst ->
	 [DomainName (str, domainlst, rrlst)]
      | DomainName (str, domainlst, rrlst)::l
      when str = (List.hd strlst) ->
	 [DomainName (str, temp domainlst (List.tl strlst), rrlst)]
      | DomainName (_,_,_)::l ->
	 temp l strlst
      | [] -> 
	 []
      in
      temp root (self#str2list str)

   (** Add an entry and his ressource record. *)
   method add str (rrlist: rrtype list) =
      match self#get str with
      | [NotFound] -> self#replace str rrlist
      | lst -> self#replace str (lst@rrlist)

   (** Replace ressource record of this entry. *)
   method replace str (rrlist: rrtype list) =
      let rec temp root strlst rrlst =
      let hd = List.hd strlst
      and tl = List.tl strlst in
      match root with
      | [] ->
	 (match tl with
	 | [] ->
	       [DomainName (hd, [], rrlst)]
	 | _ ->
	       [DomainName (hd, temp [] tl rrlst, [])])
      | lst ->
	 (let test = ref true
   	 and recnull strlst a b c rrlst=
	    if strlst = []
	    then
	       DomainName (a, b, rrlst)
	    else
	       DomainName (a, temp b tl rrlst, c)
	 in
	 let tlst = List.map 
	    (fun (DomainName (a, b, c)) ->
	       match a = hd with
	       | false -> DomainName (a,b,c)
	       | true -> test := false; recnull tl a b c rrlst
	       ) lst in
	 match !test with
	 | false ->
	    tlst
	 | true ->
	    lst@(temp [] strlst rrlst))
      in 
      root <- temp root (self#str2list (String.lowercase str)) rrlist

   (** Check if this we have authoritative data for the domain
       {i str}. *)
   method is_authoritative (str: string) = 
      let rec temp root strlst =
      match root with
      | DomainName (str, domainlst, rrlst)::l ->
         let hd = List.hd strlst 
         and tl = List.tl strlst in
         (match str = hd with
         | true ->
            if tl = []
            then 
               true
            else
               true&&(temp domainlst tl)
         | false -> temp l strlst)
      | _ -> false
                    
      in
      temp root (self#str2list (String.lowercase str))
   
   (** Extract domain from {i str}. *)
   method get_domain str =
      try
      let res = Str.search_forward (Str.regexp "\.\(.*\)$") str 0 in 
      Str.matched_group 1 str;
      with
      | Not_found -> ""

    method private build_list lst =
	let rec temp lst =
	match lst with
	| [] -> ""
	| a::l -> (let test = temp l in
		  if test = ""
		  then
		    a
		  else
		    a^test
		  )
	and temp2 lst =
	match lst with
	| [] -> []
	| a::l -> (let test = temp l in
		  if test = ""
		  then
		    [a]@(temp2 l)
		  else
		    [a^"."^test]@(temp2 l)
		  )
	in
	temp2 lst

   method private str2list str =
     List.rev (Str.split (Str.regexp "\.") str)

end;;


(** Represents a dns header. *)
class dns_data_header =
object (self)

   val mutable id = -1
   val mutable flags = 0
   val mutable nq = -1 
   val mutable nr = -1
   val mutable nauthrr = -1
   val mutable naddrr = -1

   method get_id = id
   method get_flags = flags
   method get_nq = nq
   method get_nr = nr
   method get_nauthrr = nauthrr
   method get_naddrr = naddrr

   method set_id aid = id <- aid land 0xFFFF
   method set_flags aflags = flags <- aflags land 0xFFFF
   method set_nq anq = nq <- anq land 0xFFFF
   method set_nr anr = nr <- anr land 0xFFFF
   method set_nauthrr anauth = nauthrr <- anauth land 0xFFFF
   method set_naddrr anaddr = naddrr <- anaddr land 0xFFFF
   (* set flags methods *)
   (* QR bit *)
   method set_qr boolean =
     let oflags = flags land 0x7FFF in
     flags <- if boolean then oflags lor 0x8000 else flags
   (* OPCODE 4 bits *)
   method set_opcode num =
      flags <- (flags land 0x87FF) lor ((num land 0xF) lsl 0xB)
   (* AA bit *)   
   method set_aa boolean =
     let oflags = flags land 0xFBFF in
     flags <- if boolean then oflags lor 0x0400 else flags
   (* TC bit *)
   method set_tc boolean =
     let oflags = flags land 0xFDFF in
     flags <- if boolean then oflags lor 0x0200 else flags
   (* RD bit *)
   method set_rd boolean = 
     let oflags = flags land 0xFEFF in
     flags <- if boolean then oflags lor 0x0100 else flags
   (* RA bit *)
   method set_ra boolean =
     let oflags = flags land 0xFF7F in
     flags <- if boolean then oflags lor 0x0080 else flags
   (* Z bits *)  
   method set_z num =
      flags <- (flags land 0xFF8F) lor ((num land 0xF) lsl 0x4)
   (* RCODE bits *)  
   method set_rcode num = 
      flags <- (flags land 0xFFF0) lor (num land 0xF)
  
   (* test flags methods *)
   method is_qr = (flags land 0x8000) = 0x8000
   method get_opcode = (flags land 0x7800) lsr 11
   method is_aa = (flags land 0x0F400) = 0x0400
   method is_tc = (flags land 0x0200) = 0x0200
   method is_rd = (flags land 0x0100) = 0x0100
   method is_ra = (flags land 0x0080) = 0x0080
   method get_z = flags land 0x0070
   method get_rcode = flags land 0xF

   (** Output the header's content as a string. *)
   method output =
   (Tools.hexa_of_int id 2)^(Tools.hexa_of_int flags 2)^
   (Tools.hexa_of_int nq 2)^(Tools.hexa_of_int nr 2)^
   (Tools.hexa_of_int nauthrr 2)^(Tools.hexa_of_int naddrr 2)

end;;

(** Represents a dns question section. *)
class dns_data_question =
object (self)

    val mutable name = "" 
    val mutable typee = 0
    val mutable classs = 0
    
    method get_name = name
    method get_type = typee
    method get_class = classs

    method set_name aname = name <- aname
    method set_type atype = typee <- atype land 0xFFFF
    method set_class aclass = classs <- aclass land 0xFFFF

    (** Output the dns question's content as a string. *)
    method output =
    (self#prepare_name name)^(Tools.hexa_of_int typee 2)^
    (Tools.hexa_of_int classs 2)

    method private prepare_name name =
	let str = Str.split (Str.regexp "\.") name in 
        let buf = Buffer.create 20 in
 	let rec temp liste =
 	match liste with
 	| a :: l -> 
	    Buffer.add_char buf (Char.unsafe_chr (String.length a));
	    Buffer.add_string buf a;
	    temp l
	| [] -> ()
        in 
    	    temp str;
	    Buffer.add_char buf (Char.unsafe_chr 0);
	    Buffer.contents buf

end;;

(** Represents a dns response section. *)
class dns_data_response =
object (self)
   inherit dns_data_question as super

   val mutable ttl = 0
   val mutable rdlength = -1
   val mutable rdata = A ""

   method get_ttl = ttl
   method get_rdlength =
      if (rdlength == -1)
      then
	 self#compute_rdlength;
      rdlength
   method get_rdata = rdata

   method set_ttl attl = ttl <- attl land 0xFFFF
   method set_rdlength ardlength = rdlength <- ardlength land 0xFFF
   method set_rdata ardata = rdata <- ardata

   (** Output the dns response's content as a string. *)
   method output = 
      if (rdlength == -1) then self#compute_rdlength;
      super#output^(Tools.hexa_of_int ttl 4)^(Tools.hexa_of_int rdlength 2)^
      (self#output_rrtype rdata)

   method private output_rrtype data =
      match data with
      | A adr -> self#prepare_address adr
      | CNAME str -> super#prepare_name str 
      | HINFO (cpu,os) ->
            (Tools.hexa_of_int (String.length cpu) 1)^cpu^
            (Tools.hexa_of_int (String.length os) 1)^os
      | MX (pref, name) -> (Tools.hexa_of_int pref 2)^(super#prepare_name name)
      | NS str -> super#prepare_name str
      | NULL str -> str
      | PTR str -> super#prepare_name str
      | SOA (m,r,s,ref,ret,ex,min) -> (super#prepare_name m)^(super#prepare_name r)^
      (Tools.hexa_of_int s 4)^(Tools.hexa_of_int ref 4)^(Tools.hexa_of_int ret 4)^(Tools.hexa_of_int ex 4)^
      (Tools.hexa_of_int min 4)
      | TXT str -> (Tools.hexa_of_int (String.length str) 1)^str 
      | _ -> "#output_rrtype: Error"

   method private prepare_address str =
      let lst = Str.split (Str.regexp "\.") str in
      let rec temp lst =
	 match lst with
	 | [] -> ""
	 | a::l -> (Tools.hexa_of_int (int_of_string a) 1)^temp l
      in
      temp lst

   method private compute_rdlength =
      let temp =
      match (self#get_rdata) with
      | A str -> 4
      | AAAA str -> 16
      | CNAME str -> String.length(self#prepare_name str)
      | HINFO (cpu,os) -> String.length(cpu)+String.length(os)+2
      | MX (prf,str) -> String.length(string_of_int prf)+
      String.length(self#prepare_name str)
      | NS str -> String.length(self#prepare_name str)
      | NULL str -> String.length(str)
      | PTR str -> String.length(self#prepare_name str)
      | SOA (m,r,s,ref,ret,ex,min) -> String.length(self#prepare_name m)+
      String.length(self#prepare_name r)+5*4
      | TXT str -> String.length(str)+1
      | _ -> -1
      in
         rdlength <- temp

end;;

(** Represents a dns packet. *)
class dns_packet =
object (self)

   inherit dns_data_header as super
   
   initializer Random.self_init ()
   
   val mutable questions = ([]: dns_data_question list)
   val mutable responses = ([]: dns_data_response list)
   val mutable authorities = ([]: dns_data_response list)
   val mutable additionnals = ([]: dns_data_response list)

   method get_questions = questions
   method get_responses = responses
   method get_authorities = authorities
   method get_additionnals = additionnals

   method add_question oq = questions <- questions@oq
   method add_response ore = responses <- responses@ore
   method add_authority oau = authorities <- authorities@oau
   method add_additionnal oad = additionnals <- additionnals@oad

   method set_header (hd: dns_data_header) =
      id <- hd#get_id;
      flags <- hd#get_flags;
      nq <- hd#get_nq;
      nr <- hd#get_nr;
      nauthrr <- hd#get_nauthrr;
      naddrr <- hd#get_naddrr
      
   method get_header =
      let tmp = new dns_data_header in
      tmp#set_id id;
      tmp#set_flags flags;
      tmp#set_nq nq;
      tmp#set_nr nr;
      tmp#set_nauthrr nauthrr;
      tmp#set_naddrr naddrr;
      tmp

   (** Output the dns packet's content as a string. *)
   method output =
      (* Header init *)
      let tempo inte set = if inte == -1 then set else inte in
      super#set_id (tempo id (Random.int 65535));
      super#set_flags (tempo flags 256);
      super#set_nq (tempo nq (List.length questions));
      super#set_nr (tempo nr (List.length responses));
      super#set_nauthrr (tempo nauthrr (List.length authorities));
      super#set_naddrr (tempo naddrr (List.length additionnals));
      super#output^
      (self#output_list (questions@
      (self#list_of_questions responses)@
      (self#list_of_questions authorities)@
      (self#list_of_questions additionnals)))

   method private list_of_questions liste =
      let rec temp lst =
      match lst with
      | q::l -> (q:>dns_data_question)::temp l
      | [] -> []
      in
         temp liste

   method private output_list liste =
      let rec temp lst =
      match lst with
      | q::l -> q#output^temp l
      | [] -> ""
      in
         temp liste

end;;

(** Stream parsers than build dns objects. *)
class dns_packet_parser =
object (self)

   val mutable table = []

   method private add_string elt = table <- table@[elt]

   method private get_name value =
      let rec temp liste inte =
      match liste with
      | (o, n, e)::l when o = inte && e = 0 -> n
      | (o, n, e)::l when o = inte -> n^"."^temp table e
      | (_, _, _)::l -> temp l inte
      | [] ->
	raise (Malformed ("#get_name: There is no name at this \ 
	offset ("^(string_of_int inte)^") !!"))
      in
         temp table value;

   (** Parse an IPv6 address. *)
   method parse_addr6 stm =
      let buf = ref ("") in
      begin
      try 
      for i=0 to 7
      do
	let one = Char.code (Stream.next stm)
	and two = Char.code (Stream.next stm)
	and sep = ref ":" in
	if i = 7
	then
	   sep := ""; 
	buf := !buf^(Printf.sprintf "%.2x%.2x" one two)^(!sep)
      done;
      with
      | Stream.Failure -> raise (Malformed ("#parse_addr6: not \
      enough data."))
      end;
      !buf

   (** Parse an IPv4 address. *)
   method parse_addr stm =
      try
      let one = Char.code (Stream.next stm)
      and two = Char.code (Stream.next stm)
      and thr = Char.code (Stream.next stm)
      and fou = Char.code (Stream.next stm) in
      (string_of_int one)^"."^(string_of_int two)^
      "."^(string_of_int thr)^"."^(string_of_int fou)
      with
      | Stream.Failure -> raise (Malformed ("#parse_addr: not \
      enough data."))

  (** Parse a 'dns encoded' name (rfc1035, section 4.1.2). *)
  method parse_name stm =
    try 
    let rec temp stm =
      let buf = Buffer.create 30
      and start = ref (Stream.count stm) 
      and num = ref (Char.code (Stream.next stm)) in
      if !num != 0
      then
         match !num lsr 6 with
         | 0 -> 
            (let rec read_name inte =
            match inte with
            | 0 ->
               let temp = String.lowercase (Buffer.contents buf) in
               let zero = Char.code (List.hd (Stream.npeek 1 stm)) in
               (match zero with
               | 0xC0 -> 
	  	  (let deuze = Char.code (List.nth (Stream.npeek 2 stm) 1) in
		  self#add_string (!start, temp, deuze))
               | 0 -> self#add_string (!start, temp, 0)
               | _ -> self#add_string (!start, temp, Stream.count stm));
               temp
            | _ -> Buffer.add_char buf (Stream.next stm); read_name (inte -1)
            in 
            let temp = (read_name !num) in
            temp^"."^(self#parse_name stm))
         | 3 -> self#get_name (Char.code (Stream.next stm))
         | nbr -> ".error("^(string_of_int nbr)^")"
      else 
	""
	in
      temp stm
      with
      | Stream.Failure -> raise (Malformed ("#parse_name: not \
      enough data."))

   (** Parse a character string (rfc1035, section 5.1).
       Not fully implemented. *)
   method parse_char_string stm =
      try
      let size = Char.code (Stream.next stm) in
      let str = ref (String.create size) in
      for i=0 to (String.length !str) -1
      do
    	 !str.[i] <- Stream.next stm;
      done;
      String.lowercase !str (* return a char_string *)
      with
      | Stream.Failure -> raise (Malformed ("#parse_char_string: \
      not enough data."))

   (** Parse a dns header. *)
   method parse_header stm =
      let ret = new dns_data_header in
      let lstf = [ ret#set_id; ret#set_flags; ret#set_nq; ret#set_nr;
      ret#set_nauthrr; ret#set_naddrr] in
      try
      let nstm = Tools.extract_stream stm 12 in 
      for i=0 to 5
      do
      let fst = Char.code (Stream.next nstm) 
      and snd = Char.code (Stream.next nstm) in
      (List.nth lstf i) (fst*256+snd);
      done;
      ret
      with
      | Tools.Toosmall inte -> raise (Malformed ("#parse_header: \
      not enough data ("^(string_of_int inte)^" bytes)."))

   (** Parse a dns question. *)
   method parse_query stm = 
      let q = new dns_data_question in
      q#set_name (self#parse_name stm);
      let lstf = [ q#set_type; q#set_class ] in
      try
      let nstm = Tools.extract_stream stm 4 in
      for i=0 to 1
      do
         let fst = Char.code (Stream.next nstm) 
         and snd = Char.code (Stream.next nstm) in
         (List.nth lstf i)(fst*256+snd)
      done;
      q (* return an dns_data_question object *)
      with
      | Tools.Toosmall inte -> raise (Malformed ("#parse_query: \
      not enough data ("^(string_of_int inte)^" bytes)."))

   (** Parse a dns response. *)
   method parse_response stm = 
      let r = new dns_data_response in
      try
      (r#set_name (self#parse_name stm);
      (* type *)
      let nstm = Tools.extract_stream stm 2 in 
      let fst = Char.code (Stream.next nstm) 
      and snd = Char.code (Stream.next nstm) in
      r#set_type (fst*256+snd);
      (* class *)
      let nstm = Tools.extract_stream stm 2 in 
      let fst = Char.code (Stream.next nstm) 
      and snd = Char.code (Stream.next nstm) in
      r#set_class (fst*256+snd);
      (* ttl *)
      let nstm = Tools.extract_stream stm 4 in 
      let fst = Char.code (Stream.next nstm) 
      and snd = Char.code (Stream.next nstm)
      and thr = Char.code (Stream.next nstm) 
      and qua = Char.code (Stream.next nstm) in
      r#set_ttl (fst*256*256*256+snd*256*256+thr*256+qua);
      (* rdlength *)
      let nstm = Tools.extract_stream stm 2 in 
      let fst = Char.code (Stream.next nstm) 
      and snd = Char.code (Stream.next nstm) in
      r#set_rdlength (fst*256+snd);
      (* rdata *)
      let nstm = stm in
      match r#get_type with
      | 1 (* A *) -> r#set_rdata (A (self#parse_addr nstm))
      | 2 (* NS *) -> r#set_rdata (NS (self#parse_name nstm))
      | 5 (* CNAME *) -> r#set_rdata (CNAME (self#parse_name nstm))
      | 6 (* SOA *) -> 
         (let get_32 stme = 
	    let nstme = Tools.extract_stream stme 4 in 
            let fst = Char.code (Stream.next nstme) 
            and snd = Char.code (Stream.next nstme)
            and thr = Char.code (Stream.next nstme)
            and fou = Char.code (Stream.next nstme)
            in (fst*256+snd)*256*256+thr*256+fou
         in
         let mname = self#parse_name nstm in
         let rname = self#parse_name nstm in
         let serial = get_32 nstm in
         let refresh = get_32 nstm in
         let retry = get_32 nstm in
         let expire = get_32 nstm in
         let min = get_32 nstm in
         r#set_rdata (SOA (mname, rname, serial, refresh, retry, expire, min)))
      | 10 (* NULL *) -> r#set_rdata (NULL (self#parse_char_string nstm))
      | 12 (* PTR *) -> r#set_rdata (PTR (self#parse_name nstm))
      | 13 (* HINFO *) -> 
            let cpu = self#parse_char_string nstm
            and os = self#parse_char_string nstm in
            r#set_rdata (HINFO (cpu, os)) 
      | 15 (* MX *)  -> 
         (let nnstm = Tools.extract_stream nstm 2 in 
	 let fst = Char.code (Stream.next nnstm) 
         and snd = Char.code (Stream.next nnstm) in
         r#set_rdata (MX ((fst*256+snd),(self#parse_name nstm)));
	 )
      | 16 (* TXT *) -> r#set_rdata (TXT (self#parse_char_string nstm))
      | 28 (* AAAA *) -> r#set_rdata (AAAA (self#parse_addr6 nstm))
      | _ -> r#set_rdata Error);
      r (* return dns_data_response *)
      with
      | Tools.Toosmall _ | Stream.Failure ->
	    raise (Malformed ("#parse_response: not enough data."))
      
    (** Parse a dns packet. *)
    method parse_packet stm =
       	let answer = new dns_packet in
      	answer#set_header (self#parse_header stm);
     	let hd = answer#get_header in
    	(* parsing query *)
   	for i=1 to hd#get_nq
  	do
    	    answer#add_question [self#parse_query stm];
	done;
   	(* parsing response *)
  	for i=1 to hd#get_nr
 	do
    	    answer#add_response [self#parse_response stm];
   	done;
  	(* parsing auth. nameservers *)
 	for i=1 to hd#get_nauthrr
	do
    	    answer#add_authority [self#parse_response stm];
   	done;
  	(* parsing add. info *)
 	for i=1 to hd#get_naddrr
	do
            answer#add_additionnal [self#parse_response stm];
	done;
	table <- [];
	answer

end;;

(** Class that permits a service abstraction *)
class virtual dns_service (aserver: string) =
object (self)
   inherit dns_packet_parser

   val mutable server = aserver
   val mutable port = 53
   val mutable timeout = 60.

   val mutable query = new dns_packet 
   val mutable answer = new dns_packet

   method get_query = query
   method get_answer = answer
   method get_server = server 
   method get_port = port
   method get_timeout = timeout

   method set_query aquery = query <- aquery
   method set_answer aanswer = answer <- aanswer
   method set_server aserver = server <- aserver
   method set_port aport = port <- aport
   method set_timeout tout = timeout <- tout

   (** Send the query packet. *)
   method send sck = 
      let oc = Unix.out_channel_of_descr sck in
      output_string oc query#output;
      flush oc

   (** Send the answer packet. *)
   method receive sck =
      let str = String.create 512 in
      let size, void = Unix.recvfrom sck str 0 512 [] in
      let stm = Stream.of_string (String.sub str 0 size) in
      answer <- self#parse_packet stm

   (** Method that will do the work *)   
   method virtual action : unit

end
;;

(** Query class like in perl Net::Dns *)
class dns_query (aserver: string) =
object (self)
   inherit dns_service aserver as super

   (** Add a query to the packet with name type and class. *)
   method add_query (q: string) (t: int) (c: int)=
      let ret = new dns_data_question in
      ret#set_name q; ret#set_type t; ret#set_class c; 
      ret#set_class 1;
      query#add_question [ret]

   (** Query the nameserver. *)
   method action =
      let sck = Unix.socket Unix.PF_INET Unix.SOCK_DGRAM 0 in
      let void = Unix.connect sck
      (Unix.ADDR_INET (Unix.inet_addr_of_string server, port)) in
      super#send sck;
      match Unix.select [sck] ([]: Unix.file_descr list) ([]: Unix.file_descr list) timeout with
      | read, _ , _ when read != []-> 
	super#receive sck
      | _, _, _ -> 
	raise TimedOut

end;;

(** Do axfr request. *)
class dns_axfr (aserver: string) =
object (self)
  inherit dns_service aserver as super

  val mutable soa = 0
  val mutable leftstream = ""
  val mutable lst = []

  method get_lst = lst

   (** List this domain. *)
   method ls (q: string) =
      let ret = new dns_data_question in
      ret#set_name (String.lowercase q);
      ret#set_type 252;
      ret#set_class 1;
      query#set_rd true;
      query#add_question [ret]

   method private parse_axfr stm =
      let answer = self#parse_packet stm in
      if answer#get_nr > 0 && (List.hd answer#get_responses)#get_type = 6
      then
         soa <- soa + 1;
      answer

   (** Do the work. *)
   method action =
      let sck = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
      let void = Unix.connect sck
      (Unix.ADDR_INET (Unix.inet_addr_of_string server, port)) in
      let oc = Unix.out_channel_of_descr sck in
      output_string oc (Tools.hexa_of_int (String.length super#get_query#output)
      2);  (* let's send the packet size ... *)
      flush oc;
      super#send sck;
      let loop = ref true in
      while !loop
      do
         (match Unix.select [sck] ([]: Unix.file_descr list) ([]: Unix.file_descr list) timeout with
         | read, _ , _ when read != []-> 
            (let str = String.create 4096 in
            let size = Unix.handle_unix_error (Unix.recv sck str 0 4096)
            [] in
            self#break_stream (String.sub str 0 size);
            (* First elt must be a SOA ... *)
            let value = if lst = [] then false else (List.hd lst)#get_rcode = 6 in
            if soa = 2 || value 
            then
               loop := false;
            )
         | _, _, _ -> 
            (Unix.shutdown sck Unix.SHUTDOWN_ALL;
            loop := false))
      done;

   method private get_domain str =
      try
      let res = Str.search_forward (Str.regexp "\.\(.*\)$") str 0 in 
      Str.matched_group 1 str;
      with
      | Not_found -> ""
      
   method private break_stream str =
      leftstream <- leftstream^str;
      let test = ref true 
      and size = ref ((String.length leftstream))in
      while !test
      do
         try
         if !size <= 3
         then
            raise Stream.Failure
         else
            begin
            let stm = Stream.of_string leftstream in 
            let one = Char.code (Stream.next stm)
            and two = Char.code (Stream.next stm) in
            let sz = one*256+two in
            (if !size-2 >= sz 
            then
               begin
               let nstm = Tools.extract_stream stm sz in
               size := !size - sz - 2;
               leftstream <- Tools.stream_to_string stm;
               try
	       let str = Tools.stream_to_string nstm in
	       let tmp1 = Stream.of_string str in
               let pkt = self#parse_axfr tmp1 in
               lst <- lst@[pkt];
	       with
               | Tools.Toosmall inte -> test := false
	       | Unknow_type str -> print_string ("Catch: "^str^"\n");
               if soa = 2
               then
                  raise Stream.Failure
               end
            else
               raise Stream.Failure)
            end;
         with 
         | Stream.Failure -> test := false;
      done
      
end;;
