(**
 dns.ml --- host command in Ocaml
                                                                         
 Copyright (C) 2002  Guillaume Valadon <guillaume@valadon.net>
                     Manu Letellier <manu.letellier@gmail.com>
                                                                         
 This program is free software; you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 2 as          
 published by the Free Software Foundation.                              
                                                                         
 This program is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranty of            
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     
 General Public License for more details.                                

*)

(** Pretty print of dns_* object *)

(** Get a {i string} that represents the ressource record type from 
    an {i integer}. *)
let type_to_string typ = 
   match typ with
   | 1 -> "A"
   | 2 -> "NS"
   | 5 -> "CNAME"
   | 6 -> "SOA"
   | 12 -> "PTR"
   | 13 -> "HINFO"
   | 15 -> "MX"
   | 16 -> "TXT"
   | 28 -> "AAAA"
   | 252 -> "AXFR"
   | 255 -> "ANY"
   | a -> raise (Dns.Unknow_type (string_of_int a))
;;

(** Give an {i integer} from a {i Dns.rrtype}. *)
let rrtype_to_int rr = 
   match rr with
   | Dns.A _-> 1
   | Dns.AAAA _-> 28 
   | Dns.AXFR -> 252
   | Dns.CNAME _ -> 5
   | Dns.HINFO (_, _) -> 13
   | Dns.MX (_, _) -> 15
   | Dns.NS _ -> 2
   | Dns.NULL _ -> 10
   | Dns.PTR _ -> 12 
   | Dns.SOA (_, _, _, _, _, _, _) -> 6
   | Dns.TXT _ -> 16
   | Dns.Error -> -1
   | Dns.NotFound -> raise (Dns.Unknow_type "Dns.NotFound is not a valid rrtype.")
;;

(** Get an {i integer} that represents the ressource record type from 
    an {i string}. *)
let string_to_type str = 
   match str with
   | "A" -> 1
   | "AAAA" -> 28
   | "ANY" | "*" -> 255
   | "AXFR" -> 252
   | "CNAME" -> 5
   | "HINFO" -> 13
   | "MX" -> 15
   | "NS" -> 2
   | "NULL" -> 10
   | "PTR" -> 12 
   | "SOA" -> 6
   | "TXT" -> 16
   | _ -> raise (Dns.Unknow_type str)
;;

(** Print a {i Dns.dns_data_header} object. *)
let print_header header =
   print_string ("ID: "^(string_of_int header#get_id)^"\n");
   print_string ("Flags: "^(string_of_int header#get_flags)^"\n");
   print_string ("Questions: "^(string_of_int header#get_nq)^"\n");
   print_string ("Responses: "^(string_of_int header#get_nr)^"\n");
   print_string ("Authority: "^(string_of_int header#get_nauthrr)^"\n");
   print_string ("Additionnal: "^(string_of_int header#get_naddrr)^"\n");
   print_string "\n";
;;

(** Print the content of dns header flags. *)
let print_flags integer =
   (* QR bit *)
   if (integer land 0x8000) == 0x8000
   then
      print_string "Response message\n"
   else
      print_string "Query message\n";
   (* OPCODE 4 bits *)
   (* Must check this ... *)
   begin
   match (integer land 0x7800) lsr 11 with
   | 0 -> print_string "Standard query\n"
   | 1 -> print_string "Inverse query\n"
   | 2 -> print_string "Server status request\n"
   | _ -> print_string "reserved for future use\n"
   end;
   (* AA bit *)
   if (integer land 0x0400) == 0x0400
   then
      print_string "Nameserver is authoritative\n"
   else
      print_string "Nameserver is not authoritative\n";
   (* TC bit *)
   if (integer land 0x0200) == 0x0200
   then
      print_string "Message was truncated\n"
   else
      print_string "Message was not truncated\n";
   (* RD bit *)
   if (integer land 0x0100) == 0x0100
   then
      print_string "Recursion is desired\n"
   else
      print_string "Recursion is not desired\n";
   (* RA bit *)
   if (integer land 0x0080) == 0x0080
   then
      print_string "Recursion is available\n"
   else
      print_string "Recursion is not available\n";
   (* Z bits *)
   if (integer land 0x0070) == 0
   then
      print_string "Z bits are 0\n"
   else
      print_string "Z bits are not 0\n";
   (* RCODE bits *)
   (match (integer land 0x000F) with
   | 0 -> print_string "No error\n"
   | 1 -> print_string "Format error - The name server was \ 
      unable to interpret the query\n"
   | 2 -> print_string "Server failure - The name server was \
      unable to process this query due to a problem with the \
      name server\n"
   | 3 -> print_string "Name Error - Meaningful only for \
      responses from an authoritative name server, this \
      code signifies that the domain name referenced in \
      the query does not exist.\n"
   | 4 -> print_string "Not Implemented - The name server \
      does not support the requested kind of query.\n"
   | 5 -> print_string "Refused - The name server refuses \
      to perform the specified operation for policy \
      reasons.  For example, a name server may not wish \
      to provide the information to the particular requester, \
      or a name server may not wish to perform a particular 
      operation (e.g., zone transfer) for particular data.\n"
   | _ -> print_string "Reserved for future use.\n");
   print_string "\n";
;;

(** Print a {i Dns.rrtype}. *)
let print_rdata rdata =
    match rdata with
    | Dns.A adr -> print_string ("A: "^adr^"\n")
    | Dns.AAAA adr -> print_string ("AAAA: "^adr^"\n")
    | Dns.AXFR -> print_string "AXFR\n"
    | Dns.HINFO (cpu, os) -> print_string ("CPU: "^cpu^"\nOS: "^os^"\n")
    | Dns.CNAME str -> print_string (str^"\n")
    | Dns.MX (prf, str) -> print_string ("\n   Preference: "^(string_of_int prf)^"\n   "); print_string (str^"\n")
    | Dns.NS str -> print_string ("NS: "^str^"\n")
    | Dns.NULL str -> print_string (str^"\n")
    | Dns.PTR name -> print_string (name^"\n")
    | Dns.SOA (m, r, s, ref, ret, ex, min) ->
      (print_string ("\n   Name: "^m^"\n");
      print_string ("   Email: "^r^"\n");
      print_string ("   Serial: "^(string_of_int s)^"\n");
      print_string ("   Refresh: "^(string_of_int ref)^"\n");
      print_string ("   Retry: "^(string_of_int ret)^"\n");
      print_string ("   Expire: "^(string_of_int ex)^"\n");
      print_string ("   Minimum: "^(string_of_int min)^"\n"))
    | Dns.TXT txt -> print_string (txt^"\n")
    | Dns.Error -> raise (Dns.Unknow_type "")
    | Dns.NotFound -> print_string "Name not found\n"
;;

(** Print a {i Dns.dns_data_question list}. *)
let rec print_query liste =
   match liste with
   | query::l ->
    print_string ("Name: "^(query#get_name)^"\n");
    print_string ("Type: "^(type_to_string query#get_type)^"\n");
    print_string ("Class: "^(string_of_int query#get_class)^"\n");
    print_string "\n";
    print_query l
   |[] -> ()
;;


(** Print a {i Dns.dns_data_resonse list}. *)
let rec print_answer liste =
   match liste with
   | answer::l ->
    print_string ("Name: "^(answer#get_name)^"\n");
    print_string ("Type: "^(type_to_string answer#get_type)^"\n");
    print_string ("Class: "^(string_of_int answer#get_class)^"\n");
    print_string ("TTL: "^(string_of_int answer#get_ttl)^"\n");
    print_string "Rdata: ";
    print_rdata answer#get_rdata;
    print_string "\n";
    print_string "\n";
    print_answer l
    |[] -> ()
;;

(** Print a {i Dns.dns_packet} object. *)
let print_packet pkt =
   let hd = pkt#get_header in
   print_header hd;
   print_flags hd#get_flags;
   if hd#get_nq != 0
   then
      print_query pkt#get_questions;
   if hd#get_nr != 0
   then
      (print_string "Responses :\n-----------\n";
      print_answer pkt#get_responses);
   if hd#get_nauthrr != 0
   then
      (print_string "Authorities :\n------------------------\n";
      print_answer pkt#get_authorities);
   if hd#get_naddrr != 0
   then
      (print_string "Additionnals :\n---------------------------\n";
      print_answer pkt#get_additionnals)
;;
